//+-----------------------------------------------------------------------------+
//| Copyleft (C) 2016 REGEO (http://www.regeo.uji.es/)
//|
//| This program is free software
//| You can redistribute it and/or modify it under the terms of the GNU General Public License,
//| as published by the Free Software Foundation.
//| 
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//+----------------------------------------------------------------------------*/

#if !defined PERIMETER_DEFINED
	#define PERIMETER_DEFINED

	class CModelPerimeter
	{
		public:
			int VertexCount;
			std::vector <POINT2D> Vertex;

			int EdgeCount;
			std::vector <long> EdgeHead;
			std::vector <long> EdgeTail;

			std::vector <std::vector <long>> EdgesInVertex;  //valences of vertices (number of edges connected to every vertex)

		CModelPerimeter();
		CModelPerimeter(const CModelPerimeter& MP);
		~CModelPerimeter();
	};


	class CGraphs
	{
		public:
			int SubgraphCount;
			std::vector <long> SubgraphForVertex;
			std::vector <long> SubgraphForEdge;
			void DetermineSubgraphs(CModelPerimeter MP);
	};


	class CCuePerimeter
	{
		private:
			//Input data
			//----------
			CModelPerimeter MP_main;  //Full model
			CModelPerimeter MP;       //Current subgraph

			void Input();

			void GetImplicitData();
			void GetSubGraph(CGraphs CG, long NumGraph,
							 std::vector <long> &ReverseMapVertices,
							 std::vector <long> &ReverseMapEdges);

			//Working data
			//------------
			bool IncludeDanglingEdges;
			double ThresholdIntersection;

			//Internal functions
			long GetFirstVertex();
			bool isDangling(long numEdge, long closeEnd, long &numVisited);
			long GetDanglingEdges(long numVertex);
			bool isTjunction(long numVertex, long numEdge, long otherEdge);
			double GetInnerAngle(POINT2D P1, POINT2D P2, POINT2D P3);
			long GetNextEdge(POINT2D Previous, long Current, long CurrentEdge, bool &dangling);
			std::vector <bool> GetEdgesToCheck(long CurrentEdge);
			bool GetCloserIntersection(POINT2D a1, POINT2D b1, std::vector <bool> CheckEdge,
									   long &InterceptingEdge, POINT2D &Intercept, long &T_vertex);
			void PerimeterByEnvelope(std::vector <long> &ReverseMapVertices,
									 std::vector <long> &ReverseMapEdges,
									 std::vector <long> &PerimeterVertices,
									 std::vector <long> &PerimeterEdges,
									 std::vector <POINT2D> &Tvertex,
									 std::vector <std::vector <long>> &Tedges);

		public:
			void GetPerimeter(std::vector<std::vector<long>> &PerimeterVertices,
							  std::vector<std::vector<long>> &PerimeterEdges,
							  std::vector<std::vector<POINT2D>> &Tjunction,
							  std::vector<std::vector<std::vector<long>>> &Tedges);
	};
#endif // PERIMETER_DEFINED