// Main.cpp
//Main flow to get symmetry planes
//from a graph-like representation of a polyhedrical shape

//Based on:
//    Company P., Varley P.A.C. and Plumed R.
//    "Perimeter detection in sketched drawings of polyhedral shapes"
//    STAG: Smart Tools and Applications in Graphics (2017)
//    A. Giachetti, P. Pingi and F. Stanco (Editors)
//

#include "stdafx.h"
#include <vector>

#include "Tools_Vector.h"
#include "Tools_Geometry.h"
#include "CuePerimeter.h"


int _tmain(int argc, _TCHAR* argv[])
{
	//Assign vectors
	std::vector<std::vector<long>> PerimeterVertices;
	std::vector<std::vector<long>> PerimeterEdges;
	std::vector<std::vector<POINT2D>> Tjunction;
	std::vector<std::vector<std::vector<long>>> Tedges;

	//Find perimeter
	//--------------
	CCuePerimeter Per;
	Per.GetPerimeter(PerimeterVertices,
					 PerimeterEdges,
					 Tjunction,
					 Tedges);

	//Print output
	//------------
	long SubgraphCount= PerimeterVertices.size();
	printf("Number of subgraphs= %3d \n", SubgraphCount);

	for(long NumGraph= 0; NumGraph<SubgraphCount; NumGraph++){
		printf("Perimeter for subgraph number %3d \n", NumGraph);

		printf("Start at Vertex %d\n", PerimeterVertices[NumGraph][0]);

		for(long i= 0; i<(long)PerimeterEdges[NumGraph].size(); i++){
			long side= PerimeterEdges[NumGraph][i];
			printf("Edge %d  --", PerimeterEdges[NumGraph][i]);

			long numV= PerimeterVertices[NumGraph][i+1];
			if(numV >= 0)
				printf("up to vertex %d\n", numV);
			else{
				printf("up to intersection or T-junction %d",-numV);
				printf(", between edges (%d, %d)", 
					Tedges[NumGraph][-numV-1][0], Tedges[NumGraph][-numV-1][1]);
				printf(", at coordinates (%f, %f)\n",
					Tjunction[NumGraph][-numV-1].x, Tjunction[NumGraph][-numV-1].y);
			}
		}
	}

	printf("\n");


	system("PAUSE");
	return 0;
}


